/**
 * $Id: ActorState.java,v 1.7 2002/08/21 09:21:14 deselt Exp $
 *
 * This is the main Actor class.
 */

package salsa.language;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

public class ActorState extends Thread implements java.io.Serializable {
   protected Mailbox mailbox;
   protected UniversalActor self;

   public ActorState() {
      mailbox = new Mailbox(this);

      setName(getClass().getName());
      start();
   }

   public ActorState(UniversalActor _self) {
      this();
      self = _self;
   }

	public void setSelf(UniversalActor _self) {
		self = _self;
	}  


   public void send(Message message) {
      RunTime.receivedMessage();
      mailbox.put(message);
   }

   /**
      The process method is responsible for determining which
      child class method should be invoked for a specific message.
      The process method first invokes the chosen method, updates any
      tokens and then sends any continuation the message contains.

      Note that Class.getMethod() is not used for method resolution.
      The getMethod() method will not match methods for which a
      parameter type is unknown. This algorithm will in the case that
      only one method name matches and has the same number of arguments.

      Instead the following algorithm is applied recursively to all
      this class and all superclasses until the ActorState class is
      reached. Null arguments are given Object parameter types.

      1.) Check for an exact match between method name and parameter
          type class names.

      2.) Check for an exact match between method name, parameter
          number and for assignability between parameter types.

      3.) If above two checks fail, repeat for superclass if ActorState
          is the superclass then fail.
   */
   synchronized void process(Message message) {

      String messageName = message.getMessageName();
      Object[] messageArguments = message.getArguments();

      // Determine the parameter types for supplied arguments.
      Class[] messageParameterTypes = new Class[ messageArguments.length ];

      for (int i=0;i<messageParameterTypes.length;++i) {

         // Check to see if the argument is null. If so set its type
         // to be object.
         if (messageArguments[i] == null) {
            messageParameterTypes[i] = new Object().getClass();

         // Otherwise get the argument's type.
         } else {
            messageParameterTypes[i] = messageArguments[i].getClass();
         }
      }


      // Check the current Class and then all super classes for a matching method.
      Class state = getClass();
      Method match = null;

      while (!state.getName().equals( "salsa.language.ActorState" )) {

         // Examine each method declared in this class.
         Method[] declared = state.getDeclaredMethods();

         for (int i=0;i<declared.length;++i) {

            Method candidate = declared[i];
            Class[] candidateParameterTypes = candidate.getParameterTypes();

            // Check for an exact match between method name and parameter number.
            if (candidate.getName().equals( messageName ) &&
                  candidateParameterTypes.length == messageParameterTypes.length) {

               // Check if all the parameter types match exactly.
               int j;
               for (j=0;j<messageParameterTypes.length;++j) {

                  if (!messageParameterTypes[j].getName().equals( candidateParameterTypes[j].getName() )) {
                     break;
                  }
               }

               // Matched all successfully.
               if (j == messageParameterTypes.length) {
                  match = candidate;
                  break;
               }
            }
         }

         // See if a match has been found at this class.
         if (match != null) {
            break;
         }

         for (int i=0;i<declared.length;++i) {

            Method candidate = declared[i];
            Class[] candidateParameterTypes = candidate.getParameterTypes();
            // Check for an exact match between method name and parameter number.
            if (candidate.getName().equals( messageName ) &&
                  candidateParameterTypes.length == messageParameterTypes.length) {

               // Check if the candidate parameter types are assignable from
               // the message parameter types.
               int j;
               for (j=0;j<messageParameterTypes.length;++j) {

                  if (!candidateParameterTypes[j].isAssignableFrom( messageParameterTypes[j] ) ) {

                     // If the candidate parameter type is a primitive type check to see if the
                     // message parameter is any of the nine primitive wrapper classes

                     if (!((candidateParameterTypes[j].isPrimitive() || candidateParameterTypes[j].isArray()) &&
                         (
                            candidateParameterTypes[j] == Integer.TYPE &&
                              messageParameterTypes[j] == Integer.class ||

									 candidateParameterTypes[j].getName().equals( "[I" ) &&
                              messageParameterTypes[j].getName().equals( "[Ljava.lang.Integer;" ) ||

                            candidateParameterTypes[j] == Boolean.TYPE &&
                              messageParameterTypes[j] == Boolean.class ||
										
									 candidateParameterTypes[j].getName().equals( "[Z" ) &&
                              messageParameterTypes[j].getName().equals( "[Ljava.lang.Boolean;" ) ||

                            candidateParameterTypes[j] == Character.TYPE &&
                              messageParameterTypes[j] == Character.class ||
										
									 candidateParameterTypes[j].getName().equals( "[C" ) &&
                              messageParameterTypes[j].getName().equals( "[Ljava.lang.Character;" ) ||

                            candidateParameterTypes[j] == Byte.TYPE &&
                              messageParameterTypes[j] == Byte.class ||
										
									 candidateParameterTypes[j].getName().equals( "[B" ) &&
                              messageParameterTypes[j].getName().equals( "[Ljava.lang.Byte;" ) ||	

                            candidateParameterTypes[j] == Short.TYPE &&
                              messageParameterTypes[j] == Short.class ||
										
									 candidateParameterTypes[j].getName().equals( "[S" ) &&
                              messageParameterTypes[j].getName().equals( "[Ljava.lang.Short;" ) ||

                            candidateParameterTypes[j] == Long.TYPE &&
                              messageParameterTypes[j] == Long.class ||
										
									 candidateParameterTypes[j].getName().equals( "[J" ) &&
                              messageParameterTypes[j].getName().equals( "[Ljava.lang.Long;" ) ||

                            candidateParameterTypes[j] == Float.TYPE &&
                              messageParameterTypes[j] == Float.class ||
										
									 candidateParameterTypes[j].getName().equals( "[F" ) &&
                              messageParameterTypes[j].getName().equals( "[Ljava.lang.Float;" ) ||

                            candidateParameterTypes[j] == Double.TYPE &&
                              messageParameterTypes[j] == Double.class ||
										
									 candidateParameterTypes[j].getName().equals( "[D" ) &&
                              messageParameterTypes[j].getName().equals( "[Ljava.lang.Double;" ) ||

                            candidateParameterTypes[j] == Void.TYPE &&
                              messageParameterTypes[j] == Void.class

                         ))) {

                        break;
                     }

							// If the matching candidate parameter is a primitive array type and
							// the message parameter is a wrapper type convert the message parameter
							// to a primitive array.
							if (candidateParameterTypes[j].isArray() &&
									messageParameterTypes[j].getName().equals( "[Ljava.lang.Integer;" )) {

									// Copy the values.
									int[] array = new int[ ((Integer[])messageArguments[j]).length ];
									for (int w=0; w < ((Integer[])messageArguments[j]).length; ++w) {
										array[w] = ((Integer[])messageArguments[j])[w].intValue();
									}
									
									messageArguments[j] = array;
							} 
							else if (candidateParameterTypes[j].isArray() &&
									messageParameterTypes[j].getName().equals( "[Ljava.lang.Boolean;" )) {

									// Copy the values.
									boolean[] array = new boolean[ ((Boolean[])messageArguments[j]).length ];
									for (int w=0; w < ((Boolean[])messageArguments[j]).length; ++w) {
										array[w] = ((Boolean[])messageArguments[j])[w].booleanValue();
									}
									
									messageArguments[j] = array;
							}
							else if (candidateParameterTypes[j].isArray() &&
									messageParameterTypes[j].getName().equals( "[Ljava.lang.Character;" )) {

									// Copy the values.
									char[] array = new char[ ((Character[])messageArguments[j]).length ];
									for (int w=0; w < ((Character[])messageArguments[j]).length; ++w) {
										array[w] = ((Character[])messageArguments[j])[w].charValue();
									}
									
									messageArguments[j] = array;
							}
							else if (candidateParameterTypes[j].isArray() &&
									messageParameterTypes[j].getName().equals( "[Ljava.lang.Byte;" )) {

									// Copy the values.
									byte[] array = new byte[ ((Byte[])messageArguments[j]).length ];
									for (int w=0; w < ((Byte[])messageArguments[j]).length; ++w) {
										array[w] = ((Byte[])messageArguments[j])[w].byteValue();
									}
									
									messageArguments[j] = array;
							}
							else if (candidateParameterTypes[j].isArray() &&
									messageParameterTypes[j].getName().equals( "[Ljava.lang.Short;" )) {

									// Copy the values.
									short[] array = new short[ ((Short[])messageArguments[j]).length ];
									for (int w=0; w < ((Short[])messageArguments[j]).length; ++w) {
										array[w] = ((Short[])messageArguments[j])[w].shortValue();
									}
									
									messageArguments[j] = array;
							}
							else if (candidateParameterTypes[j].isArray() &&
									messageParameterTypes[j].getName().equals( "[Ljava.lang.Long;" )) {

									// Copy the values.
									long[] array = new long[ ((Long[])messageArguments[j]).length ];
									for (int w=0; w < ((Long[])messageArguments[j]).length; ++w) {
										array[w] = ((Long[])messageArguments[j])[w].longValue();
									}
									
									messageArguments[j] = array;
							}
							else if (candidateParameterTypes[j].isArray() &&
									messageParameterTypes[j].getName().equals( "[Ljava.lang.Float;" )) {

									// Copy the values.
									float[] array = new float[ ((Float[])messageArguments[j]).length ];
									for (int w=0; w < ((Float[])messageArguments[j]).length; ++w) {
										array[w] = ((Float[])messageArguments[j])[w].floatValue();
									}
									
									messageArguments[j] = array;
							}
							else if (candidateParameterTypes[j].isArray() &&
									messageParameterTypes[j].getName().equals( "[Ljava.lang.Double;" )) {

									// Copy the values.
									double[] array = new double[ ((Double[])messageArguments[j]).length ];
									for (int w=0; w < ((Double[])messageArguments[j]).length; ++w) {
										array[w] = ((Double[])messageArguments[j])[w].doubleValue();
									}
									
									messageArguments[j] = array;
							} 

                  }
               }

               // Matched all successfully.
               if (j == messageParameterTypes.length) {
                  match = candidate;
                  break;
               }
            }

         }

         // See if a match has been found at this class.
         if (match != null) {

            break;
         }

         // Examine the parent class next.
         state = state.getSuperclass();
      }

      // If a match was found attempt to invoke the method, otherwise generate an exception.
      Object token = null;
      if (match != null) {

         try {
            token = match.invoke( this, messageArguments );

         } catch (InvocationTargetException e) {

            Throwable wrapped = e.getTargetException();

            System.err.println( "Invocation of " + messageName + " generated a " +
                                wrapped.getClass().getName() + " exception" + wrapped.getMessage() );

            wrapped.printStackTrace();

         } catch (IllegalAccessException e) {

            System.err.println( "Illegal access exception generated by the invocation of " + messageName +
                                " " + e.getMessage() );

            e.printStackTrace();
         }

      } else {

         String detail = "Could not find a matching method in " + getClass().getName() +
                         " and its super classes for " + messageName + "(";

         for (int i=0;i<messageParameterTypes.length;++i) {

            detail += " " + messageParameterTypes[i].getName();
         }

         // throw new NoSuchMessageException( detail + " )" );
         System.err.println( detail + " )" );
      }

      // Send the continuation.
      Message continuation = message.getContinuation();
      if (continuation != null) {

         // Update tokens.
         continuation.updateTokens( token );

         continuation.getTarget().send( continuation );
      }

      RunTime.finishedProcessingMessage();
   }

   public void run() {
      while (true) {
         Message nextMessage = mailbox.get();

         // Tries to get Actor's lock, only after getting the message.
         process(nextMessage);
      }
   }

   protected synchronized void notifyEmptyMailbox() {
      notify();
   }

   public String toString() {
      return getClass().getName()+":"+super.toString();
   }
}
