/**
 * A message group contains a set of messages, which are to be sent
 * asynchronously to a group of actors, an array of tokens returned
 * by those actors after processing the messages, and a continuation 
 * to be used after all messages have been processed.
 */

package salsa.language;

import java.lang.reflect.Array;

public class JoinDirector extends UniversalActor {

	public UniversalActor selfReference;

	public class State extends UniversalActorState {
	
		private Message[] messages;
		private Object[] tokens;
		private Message continuation;
		private int tokensSet;
		
		State(Message[] messages, Message continuation) {
			this.messages = messages;
			this.tokens = new Object[messages.length];
			this.continuation = continuation;
			this.tokensSet = 0;
			
			this.self = selfReference;
		}
		
		State(Actor sender, Actor[] targets, String[] methodNames, Object[][] arguments) {

			messages = new Message[ targets.length ];

			// Create message array.
			for (int i=0;i<targets.length;++i) {

				messages[i] = new Message( sender, targets[i], methodNames[i], arguments[i] );
			}
			
			// Setup the other values.
			this.tokens = new Object[messages.length];
			this.continuation = null;
			this.tokensSet = 0;
			
			this.self = selfReference;						
		}

		public void process() {
		
			Message continuation = null;
		
			// Send each message.
			for (int i = 0; i < messages.length; i++) {
				// messages[i].getTarget().send(messages[i])
				//  -> ack(i, token);
				
				{
					//  -> ack(i, token)
					Actor[] continuationTargets = { self };
					String[] continuationMethodNames = {"ack"};
					Object[][] continuationArguments = {{ new Integer(i), new Token() }};

					continuation = new Message( self, continuationTargets, continuationMethodNames, continuationArguments );
				} 	
				
				// Add the continuation.
				messages[i].addContinuation( continuation );
				
				// Send the message.
				messages[i].getTarget().send(messages[i]);
			}
		}

		public void ack(Integer i, Object token) {
		
			// Add the token to the token array.
			tokens[i.intValue()] = token;
			
			// Check to see if all the messages have returned.
			if ( (++tokensSet) == tokens.length ) {
			
				// Check to see if there is a continuation.
				if (continuation != null) {
			
					// Check to see if tokens may be cast to a more specific type.
					Object[] specificTokens = (Object[])Array.newInstance(getReturnType(),tokens.length);
					
					for (int j=0; j<tokens.length;++j) {
						specificTokens[j] = tokens[j];
						
					}		
					
					continuation.updateTokens( specificTokens );
				
					continuation.getTarget().send( continuation );
				}
			} 
		}

		public void setContinuation(Message continuation) {
			this.continuation = continuation;
		}

		/**
	 	* If the message group is homogeneous, getReturnType returns 
	 	* an array of such homogeneous return type.  Otherwise,
	 	* getReturnType returns an array of Object type.
	 	*/
		public Class getReturnType() {

			if (tokens.length == 0) {
				return Object.class;
			}

			Class defaultReturnType = tokens[0].getClass();
			
			for (int i=1; i<tokens.length; i++) {
				if (defaultReturnType != tokens[i].getClass()) {
					defaultReturnType = Object.class;
					break; // return array of Objects type.
				}
			}
			
			// All methods return VOID.
			if (defaultReturnType == Void.TYPE) {
				defaultReturnType = Object.class;
			}
				
			return defaultReturnType;
		}
		
	}
	
	public JoinDirector(Message[] messages, Message continuation) {
	
		selfReference = this;
		actorState = new State(messages, continuation);
	}
	
	public JoinDirector(Actor sender, Actor[] targets, String[] methodNames, Object[][] arguments) {
	
		selfReference = this;
		actorState = new State(sender, targets, methodNames, arguments);
	}
	
	/*
	public Class getReturnType(  ) {
		
		return ((State)actorState).getReturnType();
		
	}
	*/
	
	public void setContinuation(Message continuation) {
	
		((State)actorState).setContinuation( continuation );
	}
	
}
