/**
 * $Id: UniversalActor.java,v 1.9 2002/08/21 09:21:16 deselt Exp $
 *
 * This is the Universal Actor class.  Universal actors can be
 * accessed remotely.  Clients need only have their UAN.
 * Universal actors can also migrate to an arbitrary remote location
 * described by a UAL.
 *
 * The host in the UAN needs to be running a UAN Daemon.  The host
 * specified in the UAL needs to be running an RMSP Daemon.
 *
 * An RMSP Daemon is responsible for keeping track of actors hosted
 * in its internal Stage.  It may also be used as a proxy to communicate
 * with actors hosted by AppletStages served by an HTTP server in the
 * same host as the RMSP daemon.
 *
 * Serialization of an actor passes an Actor Reference, i.e. an equivalent
 * actor with its "local" flag set to false.  The actor being serialized
 * remains in the current stage.  So, actors are passed BY REFERENCE.
 *
 * Migration of an actor moves the Actor to a new Stage, leaving behind
 * an Actor Reference (forwarder).  The process also updates the UAN
 * daemon corresponding to the actor's name with the new location.
 * All internal actor references are serialized (passed by reference).
 *
 * Anonymous actors are allowed (no UAN), as well as named actors with
 * no current location (UAL).  Actors with no UAN and no UAL should
 * not be Universal.
 *
 * Author: Carlos Varela   cvarela@uiuc.edu   1999/11/13.
 */

package salsa.language;

import java.io.*;
import java.net.*;
import wwc.messaging.*;
import wwc.naming.*;

public class UniversalActor extends Actor {
	protected UAN uan;
	protected UAL ual;
	protected boolean local;

	/**
		This is a reference to the theater the actor is currently executing in.
		it is used to discover existing Actor References.
	*/
	protected transient Theater theater;

	public UAN getUAN() {
		return uan;
	}

	public UAL getUAL() {
		return ual;
	}

	public UniversalActor() {
		super();
		uan = null;
		ual = null;
		local = true;
	}


// for testing purposes
       	protected void finalize() {
               	//we use the finalize call to cleanup the actor's assicoated objects
				 // System.out.println("finializing an actorReference: " + this.toString());

               	try {
                       	super.finalize();
               	} catch (Throwable t) {
                       	System.err.println("Error finalizing actor: "+t);
        	}
	}

	public void send(Message message) {
		if (actorState != null) {
			// local actor:
			//System.out.println("Local send: "+message);

			super.send(message);
		} else {
			// remote actor:
			try {
				//System.out.println("Remote send:"+message);
				Socket s = new Socket(ual.getHost(), ual.getPort());
				ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(s.getOutputStream()));
				out.writeObject(message);
				out.flush();
				out.close();
				s.close();
			} catch (UnknownHostException e) {
				System.err.println("Error sending message "+message+":"+e);
			} catch (IOException e) {
				System.err.println("Error sending message "+message+":"+e);
			}
		}
	}

// System environment actors (standardOutput, standardInput, etc)
// should not be serialized at run-time... they're static
// variables:  therefore they don't get serialized.  Need to be
// dynamically bound in the new stage though... not really, since
// new static actors get created at remote stage.
// System environment actors (standardOutput, standardInput, etc)
// should be dynamically bound at "read" time...

        private void writeObject(java.io.ObjectOutputStream out) throws IOException {
                if ((uan == null) && (ual == null)) {
                        // for now, just print a Warning:
                        System.out.println("Warning: Trying to write a nonUniversal actor: "+this);
                }

                out.writeObject(uan);
                out.writeObject(ual);

                out.flush();
        }

        private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
                uan = (UAN) in.readObject();
                ual = (UAL) in.readObject();

                local = false;
        }

	public void resolveActorState(Theater newTheater) {

		// Check to see if the current theater has already been resolved.
		if (theater == null) {

			theater = newTheater;

			// if local in registry, replace "this" to *local* Actor.
			String localHost = null;
			try {
				localHost = InetAddress.getLocalHost().getHostName();

			} catch (UnknownHostException e) {
				localHost = null;
			}

			if ((ual != null) && (ual.getHost().equals(localHost)) &&
			    (ual.getPort() == theater.getPort())) {

				// Reading a *local* actor.
				local = true;
				if (theater.getActorUAL(ual.getIdentifier()) != null) {
					actorState = theater.getActorUAL(ual.getIdentifier()).actorState;
				}
			}

		}
	}
}
