/**
 * $Id: UniversalActorState.java,v 1.4 2002/08/21 09:21:16 deselt Exp $
 *
 * This is the Universal Actor class.  Universal actors can be
 * accessed remotely.  Clients need only have their UAN.
 * Universal actors can also migrate to an arbitrary remote location
 * described by a UAL.
 *
 * The host in the UAN needs to be running a UAN Daemon.  The host
 * specified in the UAL needs to be running an RMSP Daemon.
 *
 * An RMSP Daemon is responsible for keeping track of actors hosted
 * in its internal Stage.  It may also be used as a proxy to communicate
 * with actors hosted by AppletStages served by an HTTP server in the 
 * same host as the RMSP daemon.
 * 
 * Serialization of an actor passes an Actor Reference, i.e. an equivalent
 * actor with its "local" flag set to false.  The actor being serialized
 * remains in the current stage.  So, actors are passed BY REFERENCE.
 * 
 * Migration of an actor moves the Actor to a new Stage, leaving behind
 * an Actor Reference (forwarder).  The process also updates the UAN
 * daemon corresponding to the actor's name with the new location.
 * All internal actor references are serialized (passed by reference).
 *
 * Anonymous actors are allowed (no UAN), as well as named actors with
 * no current location (UAL).  Actors with no UAN and no UAL should
 * not be Universal.
 *
 * Author: Carlos Varela   cvarela@uiuc.edu   1999/11/13.
 */

package salsa.language;

import java.io.*;
import java.net.*;
import wwc.messaging.*;
import wwc.naming.*;

public class UniversalActorState extends ActorState {
	public UniversalActorState() {
		super();
	}

	public UniversalActorState(UniversalActor _self) {
		super( _self );
	}

	public void setSelf(UniversalActor _self) {
		super.setSelf( _self );
	}

	/**
	 * Binds this object to a UAN and UAL so it can be located later.
	 *
	 * @param uan The universal actor name to bind this actor to.
	 * @param ual The universal actor locator to bind this actor to.
	 */
	public synchronized void bind(String uan, String ual) throws UANException, UALException {
		bind(new UAN(uan), new UAL(ual));
	}

	/**
	 * Binds this object to a UAN and UAL so it can be located later.
	 *
	 * @param uan The universal actor name to bind this actor to.
	 * @param ual The universal actor locator to bind this actor to.
	 */
	public synchronized void bind(UAN uan, UAL ual) throws UANException, UALException {
		self.uan = uan;
		self.ual = ual;
		try {
			// (1) moves "this" to UAL.
			moveTo(self.ual);
			// (2) update UANDaemon with new UAL.
			if (self.uan != null)  // not anonymous
				uan.updateLocation(ual);
		} catch (UnknownHostException e) {
			self.uan = null;
			self.ual = null;
			throw new UALException("RMSP Server "+ual.getHost()+" unreachable: "+e);
		} catch (IOException e) {
			self.uan = null;
			self.ual = null;
			throw new UALException("java.io.IOException: "+e);
		} catch (UANIOException e){
			self.uan = null;
			throw new UANException("Actor Name Server "+uan.getHost()+" unreachable: "+e);
		}
		self.local = false;
	}

	/** 
	 * These methods require the calling Actor to have the same type as 
	 * the actor corresponding to the UAN given.  
	 * For examples, see migration.Agent, migration.Migrate and
	 * migration.RemoteSend.
	 */
	public void getReferenceByName(String uan) throws UANException {
		getReferenceByName(new UAN(uan));
	}

	public void getReferenceByName(UAN uan) throws UANException {
		self.uan = uan;
		self.ual = uan.getLocation();
		self.local = false;
	}

	public void getReferenceByLocation(String ual) throws MalformedUALException {
		getReferenceByLocation(new UAL(ual));
	}

	public void getReferenceByLocation(UAL ual) {
		self.uan = null;
		self.ual = ual;
		self.local = false;
	}

   public UAN getUAN() {
   	return self.getUAN();
   }
                        
   public UAL getUAL() {
   	return self.getUAL();
   }

	synchronized void process(Message message) {
		if (!self.local) // Actor has moved with its mailbox!
			RunTime.finishedProcessingMessage();
		else
			super.process(message);
	}

	public synchronized void migrate(String newUAL) throws UALException, MigrationException {
		migrate(new UAL(newUAL));
	}

	public synchronized void migrate(UAL newUAL) throws MigrationException {
		if (!self.local)
			throw new MigrationException("Migration to "+newUAL+" failed:  Actor is not in this stage!");
		try {
			bind(self.uan, newUAL);
		} catch (Exception e) {
			throw new MigrationException(e.getMessage());
		}
	}

	void moveTo(UAL ual) throws UnknownHostException, IOException {
		Socket s = new Socket(ual.getHost(), ual.getPort());
		ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(s.getOutputStream()));

		MigrationWrapper migrationWrapper = new MigrationWrapper(self,this);

		out.writeObject(migrationWrapper);
		out.flush();
		out.close();
		s.close();

		self.actorState = null;
		self.local = false;
		//set the state to garbage
//		actorState.isGarbage();
	}


        private void writeObject(java.io.ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
		out.flush();
        }
                 
        private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
        	in.defaultReadObject();
	}

}
