/**
 * $Id: SalsaCodeGenerator.java,v 1.11 2002/08/21 09:21:27 deselt Exp $
 *
 * This is the Salsa programming language code generator.
 * We use the abstract syntax tree produced by the parsing phase to build
 * Java files corresponding to the original Salsa files being compiled.
 *
 * Author: Carlos Varela   cvarela@uiuc.edu   1999/09/14.
 *
 */
import java.io.*;

public class SalsaCodeGenerator {
    SalsaNode tree;
    PrintWriter writer;
    static String behaviorName;
    static String extendedBehaviorName;
    static boolean generateMain = false;
    static String indent;

    SalsaCodeGenerator(SalsaNode node, OutputStream stream) {
   this.tree = node;
   this.writer = new PrintWriter(stream);
   indent = "\t\t";
    }

    public void generateCode() throws CodeGenerationException {
   tree.generateJavaCode(writer);
   writer.flush();
   writer.close();
    }

    public static String generateMessageSend(MessageSendingData msData) {
   return generateMessageSend(msData.getTarget(),
               msData.getTargets(),
               msData.getMethodNames(),
               msData.getArguments(),
               msData.getTokens(),
               msData.getSender(),
               true,
               msData.getTokenClass());
    }

    public static String generateMessageSend(MessageSendingData msData, boolean withSend) {
   return generateMessageSend(msData.getTarget(),
               msData.getTargets(),
               msData.getMethodNames(),
               msData.getArguments(),
               msData.getTokens(),
               msData.getSender(),
               withSend,
               msData.getTokenClass());
    }

    public static String generateMessageSend(String target,
                    String targets,
                    String methodNames,
                    String arguments,
                    String tokens,
                    String sender) {
   return generateMessageSend(target,
               targets,
               methodNames,
               arguments,
               tokens,
               sender,
               true,
               null);
    }

    public static String generateMessageSend(String target,
                    String targets,
                    String methodNames,
                    String arguments,
                    String tokens,
                    String sender,
                    boolean withSend) {
   return generateMessageSend(target,
               targets,
               methodNames,
               arguments,
               tokens,
               sender,
               withSend,
               null);
    }

    public static String generateMessageSend(String target,
                    String targets,
                    String methodNames,
                    String arguments,
                    String tokens,
                    String sender,
                    boolean withSend,
                    String tokenClass) {


		if (withSend) {
      	return indent + "{\n" +
                indent + "\tActor[] _targets = { " + targets + " };\n" +
                indent + "\tString[] _methodNames = { " + methodNames + " };\n" +
                indent + "\tObject[][] _arguments = { " + arguments + " };\n\n" +
                indent + "\t" + target + ".send( " +
                "new Message( " + target + ", _targets, _methodNames, _arguments ) );\n" +
                indent + "}\n";
		} 
		else {
			return   indent + "\n" +
               	indent + "\tActor[] _targets = { " + targets + " };\n" +
               	indent + "\tString[] _methodNames = { " + methodNames + " };\n" +
               	indent + "\tObject[][] _arguments = { " + arguments + " };\n\n" +
               	indent + "\tMessage " + target + " = " +
               	"new Message( " + sender + ", _targets, _methodNames, _arguments );\n" +
               	indent + "\n";
		}


    }

    public static void setBehaviorName(String behavior) {
      behaviorName = behavior;
    }

    public static String generateMainMethod() {
   // cvarela: create main that reports no "act" found at run-time,
   //          or give a compiler warning...
   if (!generateMain)
       return "";

   indent = "\t";
   StringBuffer mainMethodCode = new StringBuffer();
   String behaviorInstance = behaviorName.substring(0,1).toLowerCase()
       +behaviorName.substring(1,behaviorName.length());
   mainMethodCode.append("\n\tpublic static void main(String args[]){\n");
   mainMethodCode.append("\t\t"+behaviorName+" _"+behaviorInstance
               +" = new "+ behaviorName+"();\n\n\t\t");
   // _behaviorName<-act(args);
   mainMethodCode.append(generateMessageSend("_"+behaviorInstance,
                    "_"+behaviorInstance,
                    "\"act\"",
                    "{ args }",
                    "-1",
                    "null"));
   mainMethodCode.append("\n\t}\n");
   return new String(mainMethodCode);
    }

    public static String generateJoinStatement(JoinStatementData jsData) throws CodeGenerationException {
   return generateJoinStatement(jsData.getSender(),
                 jsData.getTarget(),
                 jsData.getTargets(),
                 jsData.getMethodNames(),
                 jsData.getTokens(),
                 jsData.getArguments(),
                 jsData.getContinuationExpression(),
                 jsData.getNumber(),
                 jsData.getWithSend(),
                 jsData.getActorArray(),
                 jsData.getWithContinuation());
    }

    public static String generateJoinStatement(String sender,
                      String target,
                      String targets,
                      String methodName,
                      String tokenPosition,
                      String arguments,
                      SalsaNode continuationExpression,
                      int number,
                      boolean withSend,
                      boolean actorArray,
                      boolean withContinuation)
   throws CodeGenerationException {
   StringBuffer joinCode = new StringBuffer();
   joinCode.append(generateJoinStatementPre(target, targets,
                   methodName, tokenPosition,
                   arguments, number, actorArray));
   joinCode.append(generateJoinStatementDirector(sender, target, number));
   joinCode.append(generateJoinStatementContinuationCode(continuationExpression, number+1, target));
   joinCode.append("\n\t\t\t"
         +target
         +".setContinuation(_continuation"
         +number
         +");\n");
   joinCode.append("\t\t}\n");
   if (withContinuation) {
       joinCode.append("\t\t//  "+target+"<-process();\n");
       joinCode.append(generateMessageSend((withSend?target:"_continuation"+(number-1)),
                  target,
                  "\"process\"", "{ }", "-1", "self",
                  withSend)
             +"\n");
   }
   return joinCode.toString();
    }

    public static String generateJoinStatementPre(String target,
                    String targets,
                    String methodNames,
                    String tokenPositions,
                    String arguments,
                    int number,
                    boolean actorArray) {
   StringBuffer joinCode = new StringBuffer();
   joinCode.append("\t\tJoinDirector "+target+" = null;\n");
   joinCode.append("\t\t{\n");
   if (actorArray) {
       joinCode.append("\t\t\tActor[] _targets"+number+" = "+targets+";\n");
       joinCode.append("\t\t\tString[] _methodNames"+number+" = new String["
             +targets+".length];\n");
       joinCode.append("\t\t\tint[] _tokenPositions"+number+" = new int["
             +targets+".length];\n");
       joinCode.append("\t\t\tObject [][] _arguments"+number+" = new Object["
             +targets+".length][];\n");
       joinCode.append("\t\t\tfor (int i = 0; i< "
             +targets+".length; i++){\n");
       joinCode.append("\t\t\t\t_methodNames"+number+"[i] = "+methodNames+";\n");
       joinCode.append("\t\t\t\t_tokenPositions"+number+"[i] = "
             +tokenPositions+";\n");
       joinCode.append("\t\t\t\t_arguments"+number+"[i] = ");
       
		 if (arguments.equals( "{}" )) {
       	joinCode.append("new Object[0];\n");
       
		 } else {
       	joinCode.append(arguments+";\n");  // may need to be more explicit.
       }
		 
		 joinCode.append("\t\t\t}\n");
		 
   } else {
       joinCode.append("\t\t\tActor[] _targets"+number+" = { "+targets+" };\n");
       joinCode.append("\t\t\tString[] _methodNames"+number+" = { "
             +methodNames+" };\n");
       joinCode.append("\t\t\tint[] _tokenPositions"+number+" = { "
             +tokenPositions+" };\n");
       joinCode.append("\t\t\tObject [][] _arguments"+number+" = { "
             +arguments+" };\n");
   }
   return joinCode.toString();
    }

    public static String generateJoinStatementContinuationCode(SalsaNode continuationExpression,
                            int number,
                            String target)
   throws CodeGenerationException {
   return continuationExpression.generateContinuationExpressionCode
       (number,
        false,
        "_continuation"+(number-1),
        target.getClass().getName()/*+".getReturnType()"*/);
    }

    public static String generateJoinStatementDirector(String sender, String director, int number) {
	 
   StringBuffer joinCode = new StringBuffer();
   joinCode.append("\t\t\t"+director+" = new JoinDirector(\n");
   joinCode.append("\t\t\t\t"+sender+".self,\n");
   joinCode.append("\t\t\t\t\t_targets"+number+",\n");
   joinCode.append("\t\t\t\t\t_methodNames"+number+",\n");
   joinCode.append("\t\t\t\t\t_arguments"+number+");\n");

   return joinCode.toString();
    }
}
