/** SALSA/World Wide Computer Project
 *
 * $Id: RMSPDaemon.java,v 1.12 2002/08/19 21:19:45 stepha Exp $
 *
 * Reference Server - Remote Message Sending Protocol Daemon
 * In order to deliver a message to a remote actor, messengers
 * reach this daemon (that can also host actors) to get the
 * reference of the target actor. The rmspd is in charge to forward
 * messengers to the target's stage (rmspd's host or remote applet).
 *
 * On entrance of an actor to a stage, the stage's internal registry
 * is updated with a pair: {relativeUAL, localReference}.  Its
 * references to local actors are updated to their current values in
 * this stage.  At this point, the actor is started locally.  If the
 * actor is an instance of Messenger, it is started with and "act()"
 * message.
 *
 * By Gregory Haik and Carlos Varela.  v0.1  June, 1999
 *
 * cvarela:  As of Salsa 0.3, the RMSPDaemon serves as a Universal
 *           Actor Stage as well.  No forwarding is thus necessary.
 *           Applet support coming later...    1999/11/16
 * Abe: 6/27/02
 * Removed ObjectInputGate thread and added debugging information. Streamlined code.
 */
package wwc.messaging;

import java.net.*;
import java.io.*;
import java.util.Hashtable;
import salsa.language.*;
import wwc.util.*;

/**
 * Contains the main method of the RMSP daemon.
 *
 * @version $I$, $G$
 * @author Gregory Haik, Carlos Varela, Abe Stephens
 */

public class RMSPDaemon implements Theater {
	// Must be a singleton so UniversalActor can get to it.
	private static RMSPDaemon instance = new RMSPDaemon();

	private final static String protocolVersion = RMSProtocol.VERSION;
	private final static String serverVersion = "rmspd/1.travis";

	private Hashtable registry;
	private ServerSocket passive;
	private int listenPort;

	public static RMSPDaemon getInstance() {
		return instance;
	}

	public Hashtable getRegistry() {
		return registry;
	}

	private RMSPDaemon() {
		registry = new Hashtable();
	}

	/**
      This method returns the current port.
   */
   	public int getPort() {
		return listenPort;
	}

   /**
      This method returns the specified UniversalActor.
   */
   	public UniversalActor getActorUAL( String relativeUAL ) {
		return (UniversalActor)registry.get(relativeUAL);
	}

   /**
      This method creates a passive socket and starts accepting connections to it.
      it contains some of the functionality once found in ObjectInputGate.
   */
	private void listen( int port ) {
		try {
			passive = new ServerSocket( port );
		} catch (IOException e) {
        		System.err.println( "IOException creating server socket." + e.getMessage() );

			return;
		}

		listenPort = passive.getLocalPort();

		System.out.println( "Theater listening on: " + listenPort );

		// Loop and accept incoming connections.
		while (true) {
			Socket incoming = null;

		         // Accept the connection.
			try {
				incoming = passive.accept();
			} catch (IOException e) {
				System.err.println( "IOException occured accepting new connection" + e.getMessage() );

				return;
			}

			//Handle the incoming connection
			ObjectHandler objectHandler = new ObjectHandler(incoming, registry, this);
			objectHandler.start();
		}
	}

	public static void main(String[] args) {
		// Handle arguments...

      		int listenPort = 4040;

		if (args.length > 0)
			if (args.length == 1)
            			if ((args[0].equals("-h")) || (args[0].equals("-?")))
               				RMSPDaemon.helpMessage();
            			else if (args[0].equals("-v"))
               				RMSPDaemon.printServerVersion();
            			else
               				RMSPDaemon.helpMessage("illegal option");
         		else if  (args.length != 2)
            			RMSPDaemon.helpMessage("too many arguments");
         		else if (!(args[0].equals("-p")))
            			RMSPDaemon.helpMessage("wrong options");
         	else
            		try {
               			listenPort = Integer.parseInt(args[1]);
            		} catch (Exception e) {
            			RMSPDaemon.helpMessage("invalid port number : " + args[1]);
            	}

         	RMSPDaemon theater = RMSPDaemon.getInstance();

         	theater.listen( listenPort );
   	}

	private static void printServerVersion() {
		System.out.println("Remote Message Sending Protocol Daemon.");
		System.out.println(serverVersion);
		System.exit(0);
	}

	private static void helpMessage(String s) {
		System.out.println(s);
		RMSPDaemon.printHelpMessage();
		System.exit(1);
	}

	private static void helpMessage() {
		RMSPDaemon.printHelpMessage();
		System.exit(0);
	}

	private static void printHelpMessage() {
		System.out.println("usage :");
		System.out.println("  java ...RMSPDaemon");
		System.out.println("  java ...RMSPDaemon -h");
		System.out.println("  java ...RMSPDaemon -v");
		System.out.println("  java ...RMSPDaemon -p portNumber");
		System.out.println("options :");
		System.out.println("  -h : print this message");
		System.out.println("  -v : print version");
		System.out.println("  -p portNumber : set the listening port to portNumber");
		System.out.println("                  default port number is 3030");
	}
}
